from time import time

import numpy as np
from scipy.interpolate import interp1d
import matplotlib.pyplot as plt


# spectral parameter
delta = 161e3  # in Hz
eta = 0
lb = 2e3  # in Hz

# correlation time
tau = [1e-7]  # in s

# acquisition parameter
acq_length = 4096
dt = 1e-6  # in s
t_echo = [0, 5e-6, 10e-6, 20e-6, 50e-6, 100e-6, 200e-6]  # all in s

# derived parameter
t_acq = np.arange(acq_length) * dt
t_max = acq_length*dt + 2*max(t_echo)
dampening = np.exp(-lb * t_acq)

# random number generator
seed = 1234
rng = np.random.default_rng(seed)

# number of random walkers
num_traj = 5000


def omega_q(delta_: float, eta_: float, theta_: float, phi_: float) -> float:
    cos_theta = np.cos(theta_)
    sin_theta = np.sin(theta_)
    return 2 * np.pi * delta_ * (3 * cos_theta * cos_theta - 1 + eta_ * sin_theta*sin_theta * np.cos(2*phi_))


def new_orientation(delta_: float, eta_: float) -> float:
    z_theta, z_phi = rng.random(2)
    theta = np.arccos(1 - 2 * z_theta)
    phi = 2 * np.pi * z_phi

    return omega_q(delta_, eta_, theta, phi)


for tau_i in tau:
    print(f'\nStart for tau={tau_i}')

    timesignal = np.zeros((acq_length, len(t_echo)))

    start = time()
    expected_jumps = round(t_max/tau_i)
    if expected_jumps > 1e7:
        print(f'Too many jumps to process, Skip {tau_i}s')
        continue

    for i in range(num_traj):
        t_passed = 0
        t = [0]
        phase = [0]
        accumulated_phase = 0

        while t_passed < t_max:
            # orientation until the next jump
            current_omega = new_orientation(delta, eta)

            # time to next jump
            t_wait = rng.exponential(tau_i)

            t_passed += t_wait
            accumulated_phase += t_wait * current_omega

            t.append(t_passed)
            phase.append(accumulated_phase)

        # convenient interpolation to get phase at arbitrary times
        phase_interpol = interp1d(t, phase)

        for j, t_echo_j in enumerate(t_echo):
            # effect of de-phasing and re-phasing
            start_amp = -2 * phase_interpol(t_echo_j)

            # start of actual acquisition
            timesignal[:, j] += np.cos(start_amp + phase_interpol(t_acq + 2*t_echo_j)) * dampening

        if (i+1) % 200 == 0:
            elapsed = time()-start
            print(f'Step {i+1} of {num_traj}', end=' - ')
            total = num_traj * elapsed / (i+1)
            print(f'elapsed: {elapsed:.2f}s - total: {total:.2f}s - remaining: {total-elapsed:.2f}s')

    timesignal /= num_traj

    # FT to spectrum
    freq = np.fft.fftshift(np.fft.fftfreq(acq_length, dt))
    spec = np.fft.fftshift(np.fft.fft(timesignal, axis=0), axes=0).real
    spec -= spec[0]

    t_echo_strings = list(map(str, t_echo))

    # plot spectra
    fig, ax = plt.subplots()
    lines = ax.plot(freq, spec)
    ax.set_title(f'RJ (tau = {tau_i}s)')
    ax.legend(lines, t_echo_strings)
    # plt.savefig(f'RJ_{tau_i}.png')

    # # save time signals and spectra
    # np.savetxt(f'rj_spectrum_{tau_i}.dat', np.c_[freq, spec], header='f\t' + '\t'.join(t_echo_strings))
    # np.savetxt(f'rj_timesignal_{tau_i}.dat', np.c_[t_acq, timesignal], header='t\t' + '\t'.join(t_echo_strings))

    plt.show()
